/*
 * pool.c
 *
 * Implementation of the pool interface for object allocation can caching
 *
 * Alex Fedosov
 * 2001-10-23 Gregory D. Benson
 *   name change: Pool to pool_t
 *   removed locking
 *   added user-supplied init and destroy callbacks
 */

#include <assert.h>
#include <limits.h>

#include "debug.h"
#include "pool.h"

pool_t *
pool_fill (int initial_size, int max_size, int alloc_incr,
           int free_threshold, unsigned int item_size,
           pool_malloc u_malloc, pool_free u_free,
           pool_func u_init, pool_func u_destroy)
{
  int i;
  pool_t *p;
  pool_item *current = NULL, *prev = NULL;

  DBG (("[pool_fill] initial: %d max: %d incr: %d free: %d\n",
        initial_size, max_size, alloc_incr, free_threshold));

  /* check parameters */
  if (initial_size <= 0 || max_size == 0 || max_size < -1 ||
      alloc_incr < -1 || free_threshold < -1)
    return NULL;

  /* create a pool */
  p = (pool_t *) u_malloc (sizeof (pool_t));
  assert (p != NULL);

  p->malloc = u_malloc;
  p->free = u_free;
  p->init = u_init;
  p->destroy = u_destroy;

  /* what happens when a negative number is cast to unsigned? */
  p->capacity = (max_size < 0) ? INT_MAX : max_size;
  p->free_limit = (free_threshold < 0) ? INT_MAX : free_threshold;
  p->alloc_incr = (alloc_incr < 0) ? initial_size : alloc_incr;
  p->item_size = item_size;

  /* allocate the initial elements */
  prev = NULL;
  for (i = 0; i < initial_size; i++) {
    current = (pool_item *) u_malloc (item_size);
    assert (current != NULL);
    if (u_init != NULL) {
      u_init (current);
    }
    current->next = prev;
    prev = current;
  }
  p->head = current;

  DBG (("[pool_fill] successfully added %d items\n", i));

  /* update the size and number of available elements */
  p->size = initial_size;
  p->available = initial_size;

  DBG (("[pool_fill] size: %d available: %d capacity: %d free limit %d alloc incr: %d\n \
		head: 0x%X\n", p->size, p->available, p->capacity,
        p->free_limit, p->alloc_incr, (int) p->head));

  return p;
}

pool_item *
pool_take (pool_t * p)
{
  int i;
  pool_item *e = NULL, *current = NULL, *prev = NULL;

  /* if pool is empty, add more elements */
  if (pool_empty (p)) {
    DBG (("[pool_take] pool exhausted\n"));
    DBG (("[pool_take] size: %d capacity: %d\n", p->size, p->capacity));

    /* see if we are below max capacity */
    if (p->size >= p->capacity) {
      DBG (("[pool_take] maximum capacity reached\n"));
      return NULL;
    }

    DBG (("[pool_take] re-filling the pool\n"));

    /* see if we are allowed to increase size */
    if (p->alloc_incr <= 0) {
      DBG (("[pool_take] not allowed to increase size\n"));
      return NULL;
    }

    prev = p->head;
    for (i = 0; i < p->alloc_incr; i++) {
      current = (pool_item *) p->malloc (p->item_size);
      if (p->init != NULL) {
        p->init (current);
      }
      current->next = prev;
      prev = current;
    }
    p->head = current;

    /* update size and availability */
    p->size += p->alloc_incr;
    p->available = p->alloc_incr;

    DBG (("[pool_take] added %d items, new size is %d\n", i, p->size));
  }

  /* now just take the first element and update the head pointer */
  e = p->head;
  p->head = p->head->next;

  /* update number of available elements */
  p->available--;

  DBG (("[pool_take] obtained available item, %d now available\n",
        p->available));

  /* don't know why, just felt like it. */
  e->next = NULL;

  return e;
}

void
pool_return (pool_t * p, pool_item * e)
{
  int i, extras;
  pool_item *current, *prev;

  /* add element back to the list */
  e->next = p->head;
  p->head = e;

  /* update number of available elements */
  p->available++;

  DBG (("[pool_return] returned an item to the pool, %d now available\n",
        p->available));

  /* see if anything needs deallocating */
  if ((extras = (p->available - p->free_limit)) > 0) {
    DBG (("[pool_return] pool has %d extra items\n", extras));

    /* need to deallocate that many elements */
    current = p->head;
    for (i = 0; i < extras; i++) {
      prev = current;
      current = current->next;
      if (p->destroy != NULL) {
        p->destroy (prev);
      }
      p->free (prev);
    }

    /* update head pointer */
    p->head = current;

    /* update pool size and availability */
    p->size -= extras;
    p->available -= extras;

    DBG (("[pool_return] deallocated %d items, new size is \
			%d\n", i, p->size));
  }
}

void
pool_drain (pool_t * p)
{
  int i;
  pool_item *current, *prev;
  pool_free u_free;

  /* see if pool is already empty */
  if (!pool_empty (p)) {
    DBG (("[pool_drain] pool not empty, has %d items\n", p->available));

    /* go through the list and free all elements */
    current = p->head;
    for (i = 0; i < p->available; i++) {
      prev = current;
      current = current->next;
      if (p->destroy != NULL) {
        p->destroy (prev);
      }
      p->free (prev);
    }

    DBG (("[pool_drain] pool drained, discarded %d items\n", i));
  }

  /* pool is now (or already) empty, free its memory */
  u_free = p->free;
  u_free (p);

  DBG (("[pool_drain] no more pool.\n"));
}

int
pool_empty (pool_t * p)
{
  return (p->head == NULL);
}
