/*
 * pool.h
 *
 * pool interface
 *
 * Alex Fedosov
 * Shaun Padden
 * 2001-10-23 Gregory D. Benson
 *   name change: Pool to pool_t
 *   added user-supplied init and destroy function callbacks
 *   removed lock from pool_t struct
 */

typedef int (*pool_func) (void *);
typedef void *(*pool_malloc) (size_t);
typedef void (*pool_free) (void *);

typedef struct pool_item
{
  struct pool_item *next;
}
pool_item;

typedef struct pool_t
{
  pool_item *head;              /* first element in pool */

  int available;                /* number of available elements */
  int size;                     /* current size, including allocated
                                   and avilable elements */
  int capacity;                 /* maximum size */
  int free_limit;               /* max number of available elements
                                   at any given time. extra ones will
                                   be discarded */
  int alloc_incr;               /* increments in which to allocate
                                   more elements */
  unsigned int item_size;       /* size of each element in bytes */
  pool_malloc malloc;           /* user-supplied malloc */
  pool_free free;               /* user-supplied free */
  pool_func init;               /* user-supplied init */
  pool_func destroy;            /* user-supplied destroy */
}
pool_t;

/*
   This function initializes the pool. Parameters are as follows:

   initial_size:	number of elements initially available.
   	must be greater than 0.

   max_size:		maximum number of elements a pool can contain,
  			including both available and allocated items.
  			The pool will not grow above this size.
     	if equal to -1, no limit on pool size.
   	if equal to 0, returns error.

   alloc_incr:		number of additional elements that will be allocated
   			once no elements are available.
	if equal to -1, initial_size will be used as increment.
   	if equal to 0, pool size will never be increased,
   	regardless of max_size.

   free_threshold:	maximum number of available elements at any given
   			time; all extra ones will be deallocated.
   	if equal to -1, no limit on available elements.
   	if equal to 0, all elements that become available will be immediately
   	discarded. (Does not apply to elements available initially.)
   
   elem_size:		size of the element that the pool will contain
   			in bytes.

Return value: NULL on error, pointer to a new pool on success.

*/

pool_t *pool_fill (int initial_size, int max_size, int alloc_incr,
                   int free_threshold, unsigned int item_size,
                   pool_malloc u_malloc, pool_free u_free,
                   pool_func u_init, pool_func u_destroy);

/* get an available element from the specified pool */
pool_item *pool_take (pool_t * p);

/* return a no-longer-needed item back to the pool */
void pool_return (pool_t * p, pool_item * e);

/* destroy the pool, deallocating all elements currently in the pool.
   Thus for proper memory management all elements taken from the pool
   should be returned before calling this function. */
void pool_drain (pool_t * p);

/* return nonzero value if pool is empty, zero otherwise */
int pool_empty (pool_t * p);
